@@ -1,12 +1,12 @@
 /*  Functions to compute k-nearest neighbours and minimum spanning trees
  *  with respect to the Euclidean metric and the thereon-based mutual
- *  reachability distances. The module provides access to a quite fast
+ *  reachability distances. The module gives access to a quite fast
  *  implementation of K-d trees.
  *
  *  For best speed, consider building the package from sources
  *  using, e.g., `-O3 -march=native` compiler flags.
  *
- *  Copyleft (C) 2025-2025, Marek Gagolewski <https://www.gagolewski.com>
+ *  Copyleft (C) 2025-2026, Marek Gagolewski <https://www.gagolewski.com>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU Affero General Public License
@@ -35,7 +35,7 @@ using namespace Rcpp;
 //' can be used by \code{\link{knn_euclid}} and \code{\link{mst_euclid}},
 //' amongst others.
 //'
-//' @param n_threads maximum number of threads to use
+//' @param n_threads maximal number of threads to use
 //'
 //' @return
 //' \code{omp_get_max_threads} returns the maximal number
@@ -81,7 +81,7 @@ int Romp_get_max_threads()
 //'
 //' @details
 //' The implemented algorithms, see the \code{algorithm} parameter, assume
-//' that \eqn{k} is rather small, say, \eqn{k \leq 20}.
+//' that \eqn{k} is rather small.
 //'
 //' Our implementation of K-d trees (Bentley, 1975) has been quite optimised;
 //' amongst others, it has good locality of reference (at the cost of making
@@ -89,12 +89,11 @@ int Romp_get_max_threads()
 //' midpoint (midrange) rule suggested by Maneewongvatana and Mound (1999),
 //' node pruning strategies inspired by some ideas from (Sample et al., 2001),
 //' and a couple of further tuneups proposed by the current author.
-//' Still, it is well-known that K-d trees
-//' perform well only in spaces of low intrinsic dimensionality.  Thus,
-//' due to the so-called curse of dimensionality, for high \code{d},
-//' the brute-force algorithm is recommended.
+//' Still, it is well-known that K-d trees perform well only in spaces of low
+//' intrinsic dimensionality.  Thus, due to the so-called curse of
+//' dimensionality, for high \code{d}, the brute-force algorithm is recommended.
 //'
-//' The number of threads used is controlled via the \code{OMP_NUM_THREADS}
+//' The number of threads is controlled via the \code{OMP_NUM_THREADS}
 //' environment variable or via the \code{\link{omp_set_num_threads}} function
 //' at runtime. For best speed, consider building the package
 //' from sources using, e.g., \code{-O3 -march=native} compiler flags.
@@ -103,19 +102,18 @@ int Romp_get_max_threads()
 //' @references
 //' J.L. Bentley, Multidimensional binary search trees used for associative
 //' searching, \emph{Communications of the ACM} 18(9), 509–517, 1975,
-//' \doi{10.1145/361002.361007}.
+//' \doi{10.1145/361002.361007}
 //'
 //' S. Maneewongvatana, D.M. Mount, It's okay to be skinny, if your friends
-//' are fat, \emph{4th CGC Workshop on Computational Geometry}, 1999.
+//' are fat, \emph{4th CGC Workshop on Computational Geometry}, 1999
 //'
 //' N. Sample, M. Haines, M. Arnold, T. Purcell, Optimizing search
 //' strategies in K-d Trees, \emph{5th WSES/IEEE Conf. on Circuits, Systems,
-//' Communications & Computers} (CSCC'01), 2001.
+//' Communications & Computers} (CSCC'01), 2001
 //'
 //'
 //' @param X the "database"; a matrix of shape \eqn{n\times d}
 //' @param k requested number of nearest neighbours
-//'     (should be rather small)
 //' @param Y the "query points"; \code{NULL} or a matrix of shape \eqn{m\times d};
 //'     note that setting \code{Y=X}, contrary to \code{NULL},
 //'     will include the query points themselves amongst their own neighbours
@@ -289,9 +287,8 @@ List knn_euclid(
 //' @description
 //' The function determines the/a(*) minimum spanning tree (MST) of a set
 //' of \eqn{n} points, i.e., an acyclic undirected connected graph whose
-//' vertices represent the points,
-//' edges are weighted by the distances between point pairs,
-//' and have minimal total weight.
+//' vertices represent the points, and edges are weighted by the distances
+//' between point pairs and have minimal total weight.
 //'
 //' MSTs have many uses in, amongst others, topological data analysis
 //' (clustering, density estimation, dimensionality reduction,
@@ -299,49 +296,44 @@ List knn_euclid(
 //'
 //' In clustering and density estimation, the parameter \code{M} plays the role
 //' of a smoothing factor; for discussion, see (Campello et al., 2015)
-//' and the references therein.  \code{M} corresponds to the \pkg{hdbscan}
-//' Python package's \code{min_samples=M-1}.
+//' and the references therein.
 //'
-//' For \eqn{M\leq 2}, we get a spanning tree that minimises the sum of
+//' For \eqn{M\leq 1}, we get a spanning tree that minimises the sum of
 //' Euclidean distances between the points, i.e., the classic Euclidean minimum
-//' spanning tree (EMST).   If \eqn{M=2}, the function additionally returns
+//' spanning tree (EMST).   If \eqn{M=1}, the function additionally returns
 //' the distance to each point's nearest neighbour.
 //'
-//' If \eqn{M>2}, the spanning tree is the smallest with respect to
+//' If \eqn{M>1}, the spanning tree is the smallest with respect to
 //' the degree-\eqn{M} mutual reachability distance (Campello et al., 2013) given by
 //' \eqn{d_M(i, j)=\max\{ c_M(i), c_M(j), d(i, j)\}}, where \eqn{d(i,j)}
 //' is the standard Euclidean distance between the \eqn{i}-th and the \eqn{j}-th point,
 //' and \eqn{c_M(i)} is the \eqn{i}-th \eqn{M}-core distance defined as the distance
-//' between the \eqn{i}-th point and its \eqn{(M-1)}-th nearest neighbour
+//' between the \eqn{i}-th point and its \eqn{M}-th nearest neighbour
 //' (not including the query point itself).
 //'
+//' Note that (Campello et al., 2013) defines the core distance as the
+//' distance to the \eqn{(M-1)}-th nearest neighbour (or the \eqn{M}-th one,
+//' but including self).
+//'
 //'
 //' @details
-//' (*) We note that if there are many pairs of equidistant points,
+//' (*) Note that if there are many pairs of equidistant points,
 //' there can be many minimum spanning trees. In particular, it is likely
 //' that there are point pairs with the same mutual reachability distances.
 //'
-//' To make the definition less ambiguous (albeit with no guarantees),
-//' internally, the brute-force algorithm relies on the adjusted distance:
-//' \eqn{d_M(i, j)=\max\{c_M(i), c_M(j), d(i, j)\}+\varepsilon d(i, j)} or
-//' \eqn{d_M(i, j)=\max\{c_M(i), c_M(j), d(i, j)\}-\varepsilon \min\{c_M(i), c_M(j)\}},
-//' where \eqn{\varepsilon} is close to \eqn{0}. |\code{mutreach_adj}|<1 selects
-//' the former formula (\eqn{\varepsilon}=\code{mutreach_adj})
-//' whilst 1<|\code{mutreach_adj}|<2 chooses the latter
-//' (\eqn{\varepsilon}=\code{mutreach_adj}±1).
+//' To make the definition unambiguous, the \code{mutreach_ties} argument
+//' indicates the preference towards connecting to farther/closer points with
+//' respect to the original metric, or having smaller/larger core distances
+//' in cases of tied distances; see (Gagolewski, 2026).  Empirically,
+//' \code{mutreach_ties="dcore_min"} and \code{mutreach_leaves="reconnect_dcore_min"}
+//' leads to MSTs with more leaves and hubs.
 //'
-//' For the K-d tree-based methods, on the other hand, \code{mutreach_adj}
-//' indicates the preference towards connecting to farther/closer
-//' points with respect to the original metric or having smaller/larger core distances
-//' if a point \eqn{i} has multiple nearest-neighbour candidates \eqn{j'}, \eqn{j''} with
-//' \eqn{c_M(i) \geq \max\{d(i, j'),  c_M(j')\}} and
-//' \eqn{c_M(i) \geq \max\{d(i, j''),  c_M(j'')\}}.
-//'
-//' Generally, the smaller the \code{mutreach_adj}, the more leaves should
-//' be in the tree (note that there are only four types of adjustments, though).
+//' The brute force method always resolves all ties, whilst, for efficiency,
+//' the K-d tree-based algorithms use this adjustment only for the first \eqn{M}
+//' nearest neighbours, so the resulting trees might be slightly different.
 //'
 //' The implemented algorithms, see the \code{algorithm} parameter, assume
-//' that \eqn{M} is rather small; say, \eqn{M \leq 20}.
+//' that \eqn{M} is rather small.
 //'
 //' Our implementation of K-d trees (Bentley, 1975) has been quite optimised;
 //' amongst others, it has good locality of reference (at the cost of making
@@ -356,7 +348,7 @@ List knn_euclid(
 //' The "dual-tree" Borůvka version of the algorithm is, in principle, based
 //' on (March et al., 2010). As far as our implementation is concerned,
 //' the dual-tree approach is often only faster in 2- and 3-dimensional spaces,
-//' for \eqn{M\leq 2}, and in a single-threaded setting.  For another
+//' for \eqn{M\leq 1}, and in a single-threaded setting.  For another
 //' (approximate) adaptation of the dual-tree algorithm to mutual
 //' reachability distances, see (McInnes and Healy, 2017).
 //'
@@ -371,7 +363,7 @@ List knn_euclid(
 //' the "brute-force" algorithm is recommended.  Here, we provided a
 //' parallelised (see Olson, 1995) version of the Jarník (1930) (a.k.a.
 //' Prim, 1957) algorithm, where the distances are computed
-//' on the fly (only once for \eqn{M\leq 2}).
+//' on the fly (only once for \eqn{M\leq 1}).
 //'
 //' The number of threads used is controlled via the \code{OMP_NUM_THREADS}
 //' environment variable or via the \code{\link{omp_set_num_threads}} function
@@ -398,34 +390,34 @@ List knn_euclid(
 //'
 //' J.L. Bentley, Multidimensional binary search trees used for associative
 //' searching, \emph{Communications of the ACM} 18(9), 509–517, 1975,
-//' \doi{10.1145/361002.361007}.
+//' \doi{10.1145/361002.361007}
 //'
 //' S. Maneewongvatana, D.M. Mount, It's okay to be skinny, if your friends
-//' are fat, \emph{4th CGC Workshop on Computational Geometry}, 1999.
+//' are fat, \emph{4th CGC Workshop on Computational Geometry}, 1999
 //'
 //' N. Sample, M. Haines, M. Arnold, T. Purcell, Optimizing search
 //' strategies in K-d Trees, \emph{5th WSES/IEEE Conf. on Circuits, Systems,
-//' Communications & Computers} (CSCC'01), 2001.
+//' Communications & Computers} (CSCC'01), 2001
 //'
 //' R.J.G.B. Campello, D. Moulavi, J. Sander, Density-based clustering based
 //' on hierarchical density estimates, \emph{Lecture Notes in Computer Science}
-//' 7819, 2013, 160–172. \doi{10.1007/978-3-642-37456-2_14}.
+//' 7819, 2013, 160–172. \doi{10.1007/978-3-642-37456-2_14}
 //'
 //' R.J.G.B. Campello, D. Moulavi, A. Zimek, J. Sander, Hierarchical
 //' density estimates for data clustering, visualization, and outlier detection,
 //' \emph{ACM Transactions on Knowledge Discovery from Data (TKDD)} 10(1),
-//' 2015, 1–51, \doi{10.1145/2733381}.
+//' 2015, 1–51, \doi{10.1145/2733381}
 //'
 //' L. McInnes, J. Healy, Accelerated hierarchical density-based
 //' clustering, \emph{IEEE Intl. Conf. Data Mining Workshops (ICMDW)}, 2017,
-//' 33–42, \doi{10.1109/ICDMW.2017.12}.
+//' 33–42, \doi{10.1109/ICDMW.2017.12}
 //'
+//' M. Gagolewski, quitefastmst, in preparation, 2026, TODO
 //'
 //'
 //' @param X the "database"; a matrix of shape \eqn{n\times d}
-//' @param M the degree of the mutual reachability distance
-//'    (should be rather small).
-//'    \eqn{M\leq 2} denotes the ordinary Euclidean distance
+//' @param M the smoothing factor a.k.a. the degree of the mutual reachability
+//'    distance; \eqn{M\leq 1} gives the ordinary Euclidean distance
 //' @param algorithm \code{"auto"}, \code{"single_kd_tree"},
 //'    \code{"sesqui_kd_tree"}, \code{"dual_kd_tree"}, or \code{"brute"};
 //'    K-d trees can only be used for \eqn{d} between 2 and 20 only;
@@ -436,21 +428,22 @@ List knn_euclid(
 //'    use \code{0} to select the default value, currently set to 32 for the
 //'    single-tree and sesqui-tree and 8 for the dual-tree Borůvka algorithm
 //' @param first_pass_max_brute_size minimal number of points in a node to
-//'    treat it as a leaf (unless it's actually a leaf) in the first
+//'    treat it as a leaf (unless it actually is a leaf) in the first
 //'    iteration of the algorithm; use \code{0} to select the default value,
 //'    currently set to 32
-//' @param mutreach_adj adjustment for mutual reachability distance ambiguity
-//'    (for \eqn{M>2}) whose fractional part should be close to 0:
-//'    values in \eqn{(-1,0)} prefer connecting to farther nearest neighbours,
-//'    values in \eqn{(0, 1)} fall for closer NNs (which is what many other
-//'    implementations provide), values in \eqn{(-2,-1)} prefer connecting
-//'    to points with smaller core distances, values in \eqn{(1, 2)} favour
-//'    larger core distances; see below for more details
+//' @param mutreach_ties adjustment for mutual reachability distance ambiguity
+//'    (for \eqn{M>1}); one of \code{"dcore_min"}, \code{"dist_max"},
+//'    \code{"dist_min"} (default), or \code{"dcore_max"}
+//' @param mutreach_leaves a way to postprocess the leaves of the computed tree;
+//'    one of \code{"keep"} (default: do nothing),
+//'    or \code{"reconnect_dcore_min"} (try reconnecting leaves to
+//'    inner vertices which have them amongst their M nearest neighbours;
+//'    prefer vertices of the smallest core distance)
 //' @param verbose whether to print diagnostic messages
 //'
 //'
 //' @return
-//' A list with two $(M=1)$ or four $(M>1)$ elements, \code{mst.index} and
+//' A list with two $(M=0)$ or four $(M>0)$ elements, \code{mst.index} and
 //' \code{mst.dist}, and additionally \code{nn.index} and \code{nn.dist}.
 //'
 //' \code{mst.index} is a matrix with \eqn{n-1} rows and \eqn{2} columns,
@@ -463,9 +456,9 @@ List knn_euclid(
 //' the indexes (lexicographic ordering of the \code{(weight, index1, index2)}
 //' triples).  For each \code{i}, it holds \code{mst_ind[i,1]<mst_ind[i,2]}.
 //'
-//' \code{nn.index} is an \eqn{n} by \eqn{M-1} matrix giving the indexes
+//' \code{nn.index} is an \eqn{n} by \eqn{M} matrix giving the indexes
 //' of each point's nearest neighbours with respect to the Euclidean distance.
-//' \code{nn.dist} provides the corresponding distances.
+//' \code{nn.dist} provides the corresponding Euclidean distances.
 //'
 //'
 //' @examples
@@ -485,11 +478,12 @@ List knn_euclid(
 // [[Rcpp::export("mst_euclid")]]
 List mst_euclid(
     SEXP X,
-    int M=1,
+    int M=0,
     Rcpp::String algorithm="auto",
     int max_leaf_size=0,
     int first_pass_max_brute_size=0,
-    double mutreach_adj=-1.00000011920928955078125,
+    Rcpp::String mutreach_ties="dist_min",
+    Rcpp::String mutreach_leaves="keep",
     bool verbose=false
 ) {
     using FLOAT = double;  // float is not faster..
@@ -502,9 +496,10 @@ List mst_euclid(
     Py_ssize_t d = (Py_ssize_t)_X.ncol();
     bool use_kdtree = true;
     FLOAT boruvka_variant = 1.5;
+    Py_ssize_t mutreach_ties_val = -2;
 
-    if (n < 1 || d <= 1)  stop("X is ill-shaped");
-    if (M < 1 || M > n-1) stop("incorrect M");
+    if (n < 1 || d <= 1)   stop("X is ill-shaped");
+    if (M < 0 || M >= n)   stop("incorrect M");
 
     if (algorithm == "auto") {
         if (2 <= d && d <= 20) {
@@ -554,6 +549,17 @@ List mst_euclid(
         stop("invalid 'algorithm'");
 
 
+    if (mutreach_ties == "dcore_min")
+        mutreach_ties_val = -2;
+    else if (mutreach_ties == "dist_max")
+        mutreach_ties_val = -1;
+    else if (mutreach_ties == "dist_min")
+        mutreach_ties_val = 1;
+    else if (mutreach_ties == "dcore_max")
+        mutreach_ties_val = 2;
+    else
+        stop("invalid 'mutreach_ties'");
+
     std::vector<FLOAT> XC(n*d);
     Py_ssize_t j = 0;
     for (Py_ssize_t i=0; i<n; ++i)
@@ -563,23 +569,43 @@ List mst_euclid(
 
     std::vector<Py_ssize_t> mst_ind((n-1)*2);    // C-order
     std::vector<FLOAT>      mst_dist(n-1);       // TODO: use out_dist
-    std::vector<Py_ssize_t> nn_ind((M==1)?0:(n*(M-1)));
-    std::vector<FLOAT>      nn_dist((M==1)?0:(n*(M-1)));
+    std::vector<Py_ssize_t> nn_ind((M==0)?0:(n*M));
+    std::vector<FLOAT>      nn_dist((M==0)?0:(n*M));
 
     if (use_kdtree)
         Cmst_euclid_kdtree(
             XC.data(), n, d, M, mst_dist.data(), mst_ind.data(),
-            (M==1)?nullptr:nn_dist.data(), (M==1)?nullptr:nn_ind.data(),
+            (M==0)?nullptr:nn_dist.data(), (M==0)?nullptr:nn_ind.data(),
             max_leaf_size, first_pass_max_brute_size, boruvka_variant,
-            mutreach_adj, verbose
+            mutreach_ties_val, verbose
         );
     else
         Cmst_euclid_brute(
             XC.data(), n, d, M, mst_dist.data(), mst_ind.data(),
-            (M==1)?nullptr:nn_dist.data(), (M==1)?nullptr:nn_ind.data(),
-            mutreach_adj, verbose
+            (M==0)?nullptr:nn_dist.data(), (M==0)?nullptr:nn_ind.data(),
+            mutreach_ties_val, verbose
         );
 
+    if (mutreach_leaves=="keep")
+        ;
+    else if (mutreach_leaves=="reconnect_dcore_min") {
+        if (M>0) {
+            Py_ssize_t mutreach_leaves_maxiter = 10;   // TODO: param
+            while (mutreach_leaves_maxiter > 0) {
+                Py_ssize_t num_changes = Cleaves_reconnect_dcore_min(
+                    n-1, n, M,
+                    mst_dist.data(), mst_ind.data(),
+                    nn_dist.data(), nn_ind.data()
+                );
+                if (num_changes <= 0) break;
+                mutreach_leaves_maxiter--;
+            }
+        }
+    }
+    else
+        stop("invalid 'mutreach_leaves'");
+
+    // generate outputs
     Rcpp::IntegerMatrix out_mst_ind(n-1, 2);
     Rcpp::NumericVector out_mst_dist(n-1);
     for (Py_ssize_t i=0; i<n-1; ++i) {
@@ -588,18 +614,18 @@ List mst_euclid(
         out_mst_dist(i)    = mst_dist[i];
     }
 
-    if (M == 1) {
+    if (M == 0) {
         return Rcpp::List::create(
             Rcpp::Named("mst.index")=out_mst_ind,
             Rcpp::Named("mst.dist") =out_mst_dist
         );
     }
     else {
-        Rcpp::IntegerMatrix out_nn_ind(n, M-1);
-        Rcpp::NumericMatrix out_nn_dist(n, M-1);
+        Rcpp::IntegerMatrix out_nn_ind(n, M);
+        Rcpp::NumericMatrix out_nn_dist(n, M);
         Py_ssize_t u=0;
         for (Py_ssize_t i=0; i<n; ++i) {
-            for (Py_ssize_t j=0; j<M-1; ++j) {
+            for (Py_ssize_t j=0; j<M; ++j) {
                 out_nn_ind(i, j)  = nn_ind[u]+1;  // 1-based indexing
                 out_nn_dist(i, j) = nn_dist[u];
                 ++u;
